/*-
 * Copyright (C) @BABOLO  2005 Nov 2
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: asn.cc,v 1.4 2005/11/12 21:12:27 babolo Exp $
 */

#define DOHALF(A) (((A) - 1) / 2)

/***************************************************/
/*     codex.asn()       */
/***************************************************/

/*    mode.len
 * e  0.4      0.3      1.4      1.3
 * 0 ___X     .__X     ___X     __X.
 * 1 __X=     ._X=     __X=     _X=.
 * 2 _X==     .X==     _X==     X==.
 * 3 X===              X===
 * 4
 *   0011     0011     0011     0011
 */
static ssize_t
av(int mode, int len, u_char *ref, u_int64_t val) {
    ssize_t e;
    int i;
    bzero(ref, UPHALF(len));
    for (e = 0; val; e++) {
        i = (val % 10) & 0x0F;
        val = (val / 10) & 0x1FFFFFFFFFFFFFFF;
        if  (e < len) {
                ref[DOHALF(len) - (e / 2 + (mode & len & 1))]
            |=  ((e ^ (mode & len)) & 1) ? ((i << 4) & 0xF0) : (i & 0x0F);
    }   }
    if  ((mode >> 1) & e & 1) {
        for (i = 0; i < DOHALF(len); i++) {
            ref[i] = ((ref[i] << 4) & 0xF0) | ((ref[i + 1] >> 4) & 0x0F);
        }
        ref[DOHALF(len)] = (ref[DOHALF(len)] << 4) & 0xF0;
    }
    return(e);
}

static ssize_t
avr(ISO8583_Pack *pack, int fno, u_int64_t val) {
    int len = pack->package->fld[fno].len;
    u_char *ref = pack->refo[fno];
    return((av(0, len, ref, val) > len) ? 0 : UPHALF(len));
}

static ssize_t
avl(ISO8583_Pack *pack, int fno, u_int64_t val) {
    int len = pack->package->fld[fno].len;
    u_char *ref = pack->refo[fno];
    return((av(1, len, ref, val) > len) ? 0 : UPHALF(len));
}

static ssize_t
avnr(ISO8583_Pack *pack, int fno, u_int64_t val) {
    int len = pack->package->fld[fno].len;
    u_char *ref = pack->refo[fno];
    u_char b[ZSIZE];
    int y = lv(0, 2, ref);
    ssize_t e;

    if  ((e = av(0, ZSIZE * 2, b, val)) > len || len > ZSIZE * 2) {
        ifBLIN_QV3(pack->flags) warnx("ISO8583_Asn: too big %llu(%d)", val, len);
        errno = EINVAL;
        e = 0;
    } else if (y && (UPHALF(e) != UPHALF(y))) {
        ifBLIN_QV3(pack->flags) warnx("ISO8583_Asn: length mismatch %llu(%d): %d<>%d", val, len, y, e);
        errno = EINVAL;
        e = 0;
    } else if (av(0, 2, ref++, (u_int64_t)e) > 2) {
        ifBLIN_QV3(pack->flags) warnx("ISO8583_Asn: too big length %u", e);
        errno = EINVAL;
        e = 0;
    } else {
        e = UPHALF(e);
        memmove(ref, &b[ZSIZE - e], e);
        e++;
    }
    return(e);
}

static ssize_t
avnl(ISO8583_Pack *pack, int fno, u_int64_t val) {
    int len = pack->package->fld[fno].len;
    u_char *ref = pack->refo[fno];
    u_char b[ZSIZE];
    int y = lv(0, 2, ref);
    ssize_t e;

    if  ((e = av(2, ZSIZE * 2, b, val)) > len || len > ZSIZE * 2) {
        ifBLIN_QV3(pack->flags) warnx("ISO8583_Asn: too big %llu(%d)", val, len);
        errno = EINVAL;
        e = 0;
    } else if (y && (UPHALF(e) != UPHALF(y))) {
        ifBLIN_QV3(pack->flags) warnx("ISO8583_Asn: length mismatch %llu(%d): %d<>%d", val, len, y, e);
        errno = EINVAL;
        e = 0;
    } else if (av(0, 2, ref++, (u_int64_t)e) > 2) {
        ifBLIN_QV3(pack->flags) warnx("ISO8583_Asn: too big length %u", e);
        errno = EINVAL;
        e = 0;
    } else {
        e = UPHALF(e);
        memmove(ref, &b[ZSIZE - e], e);
        e++;
    }
    return(e);
}

static ssize_t
atrl(ISO8583_Pack *pack, int fno, u_int64_t val) {
    int len = pack->package->fld[fno].len;
    u_char *ref = pack->refo[fno];
    int i = len;
    for (; i--; val >>= 8) {
        ref[i] = val & 0xFF;
    }
    return(len);
}

static ssize_t
expnt(ISO8583_Pack *pack, int fno, u_int64_t val, int mode) {
    int len = pack->package->fld[fno].len;
    u_char *ref = pack->refo[fno];
    struct tm tm;
    ssize_t e = len / 2, l = len / 2 - 1;
    time_t t = (time_t)val;
    int i = 0;
#define ONUM(A, B) \
    {   if  (av(0, 2, ref++, (A)) > 2) {                                                              \
            ifBLIN_QV3(pack->flags) warnx("ISO8583_Asn: too big " B " %u", (A));                      \
            errno = EINVAL;                                                                           \
            e = 0;                                                                                    \
            break;                                                                                    \
        }                                                                                             \
        if  (++i > l) break;                                                                          \
    }

    if  (!len && (len & 1)) mode = 111;
    bzero(&tm, sizeof(struct tm));
    gmtime_r(&t, &tm);
    switch (mode) {
    case  0: ONUM(tm.tm_year / 100 + 19, "century");
    case  2: ONUM(tm.tm_year % 100     , "year"   );
    case  4: ONUM(tm.tm_mon + 1        , "month"  );
    case  6: ONUM(tm.tm_mday           , "day"    );
    case  8: ONUM(tm.tm_hour           , "hour"   );
    case 10: ONUM(tm.tm_min            , "minutes");
    case 12: ONUM(tm.tm_sec            , "secunds");
             break;
    default:
             ifBLIN_QV3(pack->flags) warnx("ISO8583_Asn: illegal mode %d", mode);
             errno = EINVAL;
             e = 0;
             break;
    case 111:
             ifBLIN_QV3(pack->flags) warnx("ISO8583_Asn: illegal length %d", len);
             errno = EINVAL;
             e = 0;
    }
    return(e);
}

static ssize_t
aim(ISO8583_Pack *pack, int fno, u_int64_t val) {
    return(expnt(pack, fno, val, 14 - pack->package->fld[fno].len));
}

static ssize_t
aat(ISO8583_Pack *pack, int fno, u_int64_t val) {
    return(expnt(pack, fno, val, 2));
}

static ssize_t
aatt(ISO8583_Pack *pack, int fno, u_int64_t val) {
    return(expnt(pack, fno, val, 8 - pack->package->fld[fno].len));
}

/* XXXX  msz */
