/*-
 * Copyright (C)2005 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2005 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: codec.c,v 1.4 2005/11/12 21:12:28 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/param.h>
#include <babolo/BLINflag.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <err.h>
#include "iso8583.h"
#ifndef ENOATTR
#define ENOATTR EINVAL
#endif
#define ZSIZE 10
#define UPHALF(A) (((A) + 1) / 2)

/*  mode  codex.len() */
#define MS_MES 0 /*                             */
#define MS_NST 1 /* 0,   ,    */

/*    codex.len() */
#define MS_NUL 0
#define MS_MOD 1
#define MS_ILL 2 /*   */
#define MS_NOT 3 /*  */

static int
kolo(int mod, int hint, int cond) {
    if  ((hint < (mod >> 1)) && (cond > (hint + (mod >> 1)))) return (-1);
    if  ((hint > (mod >> 1)) && (cond < (hint - (mod >> 1)))) return (1);
    return(0);
}

#include "val.cc"
#include "ref.cc"
#include "asn.cc"
#include "cpy.cc"
#include "len.cc"

static const struct {
    ssize_t   (*len)(ISO8583_Pack *pack, int fno, int mode);
    u_int64_t (*val)(ISO8583_Pack *pack, int fno);
    char     *(*ref)(ISO8583_Pack *pack, int fno);
    int       (*asn)(ISO8583_Pack *pack, int fno, u_int64_t val);
    int       (*cpy)(ISO8583_Pack *pack, int fno, u_char *vly);
} codex[] = {
 { eqb, ltrl, eqref, atrl, NULL} /* B_LITERAL  */
,{ eqa, NULL, eqref, NULL, cpyc} /* I_CHAR     */
,{  ll, NULL, llref, NULL, cpyl} /* B_LLCHAR   */
,{ lll, NULL,  NULL, NULL, NULL} /* B_LLLCHAR  */
,{ eqn,  lvr,  NULL,  avr, NULL} /* B_NUMERIC  */
,{ eqn,  lvl,  NULL,  avl, NULL} /* B_NUMERICl */
,{ lln, lvnr,  NULL, avnr, NULL} /* B_LLNUM    */
,{ lln, lvnl,  NULL, avnl, NULL} /* B_LLNUMl   */
,{ eqd,  tim,   dim,  aim, NULL} /* B_TIME     */
,{ eqd,  dat,   dad,  aat, NULL} /* B_DATE     */
,{ eqd, datt,  dadt, aatt, NULL} /* B_DATET    */
,{unkn, NULL,  NULL, NULL, NULL} /* B_BINARY   */
,{unkn, NULL,  NULL, NULL, NULL} /* B_BINARY   */
,{unkn, NULL,  NULL, NULL, NULL} /* B_BINARY   */
};

static ISO8583_Pack *
mark(ISO8583_Pack **pack, int fno) {
    ssize_t l;
    if  ((*pack)->package->fld[fno].convertor >= ISO8583_E_MAX) {
        ifBLIN_QV2((*pack)->flags) warnx("ISO8583_Mark: invalid package convertor %d\n", fno);
        errno = EINVAL;
        free(*pack);
        *pack = NULL;
    } else {
        if  ((l = (codex[(*pack)->package->fld[fno].convertor].len)(*pack, fno, MS_MES)) <= 0) {
            errno = EDOM;
            switch(-l) {
            case MS_NUL:
                ifBLIN_QV2((*pack)->flags) warnx("ISO8583_Mark: null member %d length\n", fno);
                break;
            case MS_ILL:
                ifBLIN_QV2((*pack)->flags) warnx("ISO8583_Mark: invalid member %d length\n", fno);
                break;
            default:
                ifBLIN_QV2((*pack)->flags) warnx("ISO8583_Mark: unverified member %d\n", fno);
                errno = ENOSYS;
            }
            free(*pack);
            *pack = NULL;
        } else if ((*pack)->msz && (*pack)->totlen + l > (*pack)->msz) {
            ifBLIN_QV2((*pack)->flags)
                warnx("ISO8583_Mark: size overload %u [%d]\n", (*pack)->msz, fno);
            errno = EOVERFLOW;
            free(*pack);
            *pack = NULL;
        } else {
            (*pack)->totlen += l;
    }   }
    return(*pack);
}

ISO8583_Pack *
ISO8583_Mark(u_int32_t flags, const ISO8583_Package *package, u_char *in, size_t msz, char *tzname) {
    int i;
    u_int64_t m;
    ISO8583_Pack *pack = NULL;

    if  (!package) {
        ifBLIN_QV2(flags) warnx("ISO8583_Mark: package absent\n");
        errno = EINVAL;
    } else {
        pack = calloc(sizeof(ISO8583_Pack), 1);
        if  (!pack) {
            ifBLIN_QV2(flags) warnx("ISO8583_Mark: calloc pack failed\n");
            errno = ENOMEM;
        } else {
            pack->flags = flags & ~ISO8583_FMS;
            pack->package = package;
            if  ((pack->tz = tzname) && !(pack->tz = strdup(tzname))) {
                ifBLIN_QV2(flags) warnx("ISO8583_Mark: strdup failed\n");
                errno = ENOMEM;
                free(pack);
                pack = NULL;
            } else if (time(&(pack->now)) < 0) {
                ifBLIN_QV2(flags) warn("ISO8583_Mark: time failed");
                free(pack);
                pack = NULL;
            } else if (in) {
                pack->msz = msz;
            ;   pack->totlen = 0;
            ;   pack->refo[0] = in;
            ;   if  (mark(&pack, 0)) {
            ;       pack->refo[1] = pack->refo[0] + pack->totlen;
            ;       m = ISO8583_Val(pack, 1) | pack->package->premask;
            ;       ifBLIN_QV3(flags) fprintf(stderr, "m=%08llX\n", m);
            ;       for (i = 1; m; i++, m <<= 1) {
            ;           if  (m & 0x8000000000000000L) {
            ;               pack->refo[i] = pack->refo[0] + pack->totlen;
            ;               if  (!mark(&pack, i)) break;
                }   }   }
            } else {
                if  (!msz) msz = ((flags & ISO8583_MSZ) == 0) ? PAGE_SIZE : (flags & ISO8583_MSZ);
            ;   pack->msz = msz;
            ;   pack->totlen = 0;
            ;   pack->flags |= ISO8583_FMS;
            ;   pack->refo[0] = calloc(msz, 1);
            ;   if  (!pack->refo[0]) {
            ;       ifBLIN_QV2(flags) warnx("ISO8583_Mark: calloc pack->refo[0] failed\n");
            ;       errno = ENOMEM;
            ;       free(pack);
                    pack = NULL;
    }   }   }   }
    return(pack);
}

void
ISO8583_Save(ISO8583_Pack *pack) {
    free(pack->tz);
    free(pack);
}

void
ISO8583_Free(ISO8583_Pack *pack) {
    if  (pack->flags & ISO8583_FMS) free(pack->refo[0]);
    ISO8583_Save(pack);
}

int
ISO8583_Test(ISO8583_Pack *pack, int fno) {
    int e = 0;

    if  (fno < pack->package->bound && pack->refo[fno]) {
        e |= ISO8583_PRES;
        if  (codex[pack->package->fld[fno].convertor].val) e |= ISO8583_VALU;
        if  (codex[pack->package->fld[fno].convertor].ref) e |= ISO8583_REFU;
    }
    return(e);
}

u_int64_t
ISO8583_Val(ISO8583_Pack *pack, int fno) {
    u_int64_t val = 0;

    if  ((ISO8583_Test(pack, fno) & ISO8583_ISVL) == ISO8583_ISVL) {
        val = (codex[pack->package->fld[fno].convertor].val)(pack, fno);
    }
    return(val);
}

char *
ISO8583_Ref(ISO8583_Pack *pack, int fno) {
    char *ref = NULL;

    if  ((ISO8583_Test(pack, fno) & (ISO8583_PRES | ISO8583_REFU)) == (ISO8583_PRES | ISO8583_REFU)) {
        ref = (codex[pack->package->fld[fno].convertor].ref)(pack, fno);
    }
    return(ref);
}

int
ISO8583_Asn(ISO8583_Pack *pack, int fno, u_int64_t val) {
    u_char *ref = pack->refo[fno];
    ssize_t e = 0;

    ifBLIN_QV4(pack->flags) fprintf(stderr, "+ISO8583_Asn [%d]=%llu\n", fno, val);
    switch (fno) {
    case 0:
        if  (!ref) {
            ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal ISO8583_Pack");
            errno = EFAULT;
        } else if (!(e = (codex[pack->package->fld[fno].convertor].asn)(pack, fno, val))) {
            ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal value [%d]=%llu", fno, val);
            errno = EINVAL;
        } else if (!pack->totlen) {
            pack->totlen += e;
        }
        break;
    case 1:
        if  (ref) {
            ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal sequence [%d]", fno);
            errno = EINVAL;
        } else if (val) {
            ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal initiation [%d]=%llu", fno, val);
            errno = EINVAL;
        } else {
            pack->refo[fno] = pack->refo[0] + pack->totlen;
            if  (!(e = (codex[pack->package->fld[fno].convertor].asn)(pack, fno, val))) {
                ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal value [%d]=%llu", fno, val);
                errno = EINVAL;
            } else {
                pack->totlen += e;
        }   }
        break;
    default:
        ifBLIN_QV4(pack->flags) fprintf( stderr, "%016llX %016llX\n"
                                       , ISO8583_Val(pack, 1)
                                       , (((u_int64_t)1 << (64 - fno + 1)) - 1)
                                       );
        if  (!ref && ((((u_int64_t)1 << (64 - fno + 1)) - 1) & ISO8583_Val(pack, 1))) {
            ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal sequence [%d]", fno);
            errno = EINVAL;
        } else {
            int n = 0;
            if  (!ref) {
                pack->refo[fno] = pack->refo[0] + pack->totlen;
                n++;
            }
            if  (!(e = (codex[pack->package->fld[fno].convertor].asn)(pack, fno, val))) {
                ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal value [%d]=%llu", fno, val);
                errno = EINVAL;
            } else if (n) {
                pack->totlen += e;
                if  (!atrl(pack, 1, (ISO8583_Val(pack, 1) | ((u_int64_t)1 << (64 - fno))))) {
                    ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: could not set bitmap");
                    errno = ENOATTR;
    }   }   }   }
    ifBLIN_QV4(pack->flags) fprintf(stderr, "-ISO8583_Asn %d totlen=%u\n", e, pack->totlen);
    return(e);
}

int
ISO8583_Cpy(ISO8583_Pack *pack, int fno, u_char *vly) {
    u_char *ref = pack->refo[fno];
    ssize_t e = 0;

    ifBLIN_QV4(pack->flags) fprintf( stderr, "+ISO8583_Cpy %016llX %016llX\n"
                                   , ISO8583_Val(pack, 1)
                                   , (((u_int64_t)1 << (64 - fno + 1)) - 1)
                                   );
    if  (!ref && ((((u_int64_t)1 << (64 - fno + 1)) - 1) & ISO8583_Val(pack, 1))) {
        ifBLIN_QV2(pack->flags) warnx("ISO8583_Cpy: Illegal sequence [%d]", fno);
        errno = EINVAL;
    } else {
        int n = 0;
        if  (!ref) {
            pack->refo[fno] = pack->refo[0] + pack->totlen;
            n++;
        }
        if  (!(e = (codex[pack->package->fld[fno].convertor].cpy)(pack, fno, vly))) {
            ifBLIN_QV2(pack->flags) warnx("ISO8583_Cpy: Illegal copy [%d]", fno);
            errno = EINVAL;
        } else if (n) {
            pack->totlen += e;
            if  (!atrl(pack, 1, (ISO8583_Val(pack, 1) | ((u_int64_t)1 << (64 - fno))))) {
                ifBLIN_QV2(pack->flags) warnx("ISO8583_Cpy: could not set bitmap");
                errno = ENOATTR;
    }   }   }
    ifBLIN_QV4(pack->flags) fprintf(stderr, "-ISO8583_Cpy %d totlen=%u\n", e, pack->totlen);
    return(e);
}

int
ISO8583_Control(u_int32_t flags, const ISO8583_Package *package) {
    int e = 0;
    if  (  package->flag != 1
        || package->fld[0].convertor != ISO8583_B_NUMERIC
        || package->fld[1].convertor != ISO8583_B_LITERAL
        || package->fld[0].len != 4
        || package->fld[1].len != 8
        ) {
        e = -1;
    }
    return(e);
}
